from werkzeug.exceptions import BadRequest
from flask import jsonify


def handle_invalid_usage(error):
    return jsonify(error.to_dict()), error.status_code


class ValidationError(BadRequest):
    """
    When invalid data is sent to via POST or PUT, this exception gets raised

    :param message: The error message
    :param field: The name of the field that raised the exception, if applicable
    :param status_code: allows overwriting of the default bad request status code, 400
    """
    status_code = 400

    def __init__(self, message, field=None, status_code=None):
        Exception.__init__(self)
        self.message = message
        if status_code is not None:
            self.status_code = status_code
        self.field = field

    def to_dict(self):
        """
        Returns a REST-friendly API response
        """
        return {self.field: self.message}


class ConfigException(Exception):
    """
    This exception gets raised if the config file is invalid
    """
    pass



from crudcast.resources import Resource


class MockApi(object):
    pass


class MockCollection(object):
    items = []

    def find(self, query):
        return MockDocument(_count=self.count)

    def insert_one(self, *args, **kwargs):
        d = MockDocument()
        self.items.append(d)
        return d

    def __init__(self, *args, **kwargs):
        self.count = kwargs.get('count', 0)


class MockMethod(object):
    def __init__(self, *args, **kwargs):
        pass


class MockApp(object):
    user_manager = None
    handle_exception = None
    handle_user_exception = None
    methods = {}

    def run(self, *args, **kwargs):
        pass

    extensions = {}
    config = {}
    crudcast_config = {
        'swagger': {
            'basePath': 'api'
        }
    }
    models = {
        'test': {
            'collection': MockCollection(),
            'fields': {}
        }
    }

    def route(self, *args, **kwargs):
        return lambda x: print(x)

    def register_blueprint(self, *args, **kwargs):
        pass

    def register_error_handler(self, *args, **kwargs):
        pass

    @classmethod
    def add_url_rule(rule, *args, **kwargs):
        pass

    def __init__(self, *args, **kwargs):
        pass

    def get_swagger_ui_view(self, *args, **kwargs):
        pass


class MockArgs(object):
    config_file = 'something invalid'
    import_name = 'Crudcast'
    port = 5000
    debug = False
    host = '0.0.0.0'
    no_load_dotenv = False
    create_admin = False


class MockParser(object):
    def parse_args(self):
        return MockArgs

    def add_argument(self, *args, **kwargs):
        pass

    def __init__(self, *args, **kwargs):
        pass


class MockDocument(object):
    _count = 0

    def items(self):
        return [{'test': 'test'}]

    def __init__(self, *args, **kwargs):
        self._count = kwargs.get('_count', 0)

    def __iter__(self):
        for item in self.items():
            yield item

    def count(self):
        return self._count

    inserted_id = '507f1f77bcf86cd799439011'


class MockModel(object):
    models = ['test', 'test2']
    count = 0
    name = None
    app = MockApp()

    def __init__(self, *args, **kwargs):
        self.name = kwargs.get('name')
        self.count = kwargs.get('count', 0)
        self.collection = MockCollection(count=self.count)
        pass

    def to_repr(self, _id):
        return [{
            'test': {}
        }]


class MockValidationError(Exception):
    pass


class MockMongo(object):
    def __getitem__(self, item):
        return {
            'test': MockCollection(),
            'test2': MockCollection(),
        }

    def __init__(self, *args, **kwargs):
        pass


class TestResource(Resource):
    def get(self):
        return {'hello': True}


from crudcast.models import Model
from crudcast.exceptions import ValidationError
from flask import abort
import bcrypt


class User(Model):
    config = {
        'username_field': 'username'
    }  # default user config

    @property
    def username_field(self):
        return self.config['username_field']

    def to_repr(self, **query):
        """
        Remove the hashed password from the response
        """
        output = super(User, self).to_repr(**query)
        new_output = []
        for user in output:
            new_user = {}
            for key, val in user.items():
                if key not in ['password', 'salt']:
                    new_user[key] = val
            new_output.append(new_user)

        return new_output

    def __init__(self, app, **options):
        for key, val in options.items():
            self.config[key] = val

        super().__init__('user', app=app)

    def user_exists(self, username):
        return self.collection.find({self.username_field: username}).count() == 1

    def check_invalid_keys(self, keys):
        key = None
        try:
            for key in keys:
                assert key in [self.username_field, 'password']
        except AssertionError:
            raise ValidationError('Invalid field', field=key)

    def validate(self, data=None, _id=None):
        if not data:
            data = {}

        if not _id:
            try:
                assert self.username_field in data.keys()
                assert 'password' in data.keys()

            except AssertionError:
                raise ValidationError('Username and password must be provided')

            if self.user_exists(data[self.username_field]):
                raise ValidationError('That user already exists')

        return data

    def hash_password(self, raw_password):
        salt = bcrypt.gensalt()
        return bcrypt.hashpw(raw_password.encode(), salt), salt

    def authenticate(self, username, password):
        if not username or not password:
            abort(401)

        try:
            users = self.find(**{self.username_field: username})
            assert users.count() == 1
            user = users[0]
            self.verify_password(username, password)
            return user

        except AssertionError:
            abort(401)

    def update(self, _id, data):
        """
        Method extended to hash the password, if a new one is provided
        """

        # the below validation is done at this level to prevent manual insertion of salt
        self.check_invalid_keys(data.keys())

        if data.get('password'):
            data['password'], data['salt'] = self.hash_password(data.pop('password'))

        return super().update(_id, data)

    def create(self, data):
        if data.get('password'):
            password = data.pop('password')
            data['password'], data['salt'] = self.hash_password(password)
        return super().create(data)

    def verify_password(self, username, password):
        users = self.find(**{self.username_field: username})
        hash = users[0]['password']
        salt = users[0]['salt']
        valid = bcrypt.hashpw(password.encode(), salt) == hash
        if not valid:
            abort(401)



from flask_restplus import Api
from crudcast.resources import ModelResource, InstanceResource, UserModelResource, UserInstanceResource


def get_api(app):
    base_path = app.crudcast_config['swagger']['basePath']
    api = Api(app, doc=False)

    mr = ModelResource
    mr.set_app(app)

    ir = InstanceResource
    ir.set_app(app)

    api.add_resource(mr, '%s/<string:model_name>/' % base_path)
    api.add_resource(ir, '%s/<string:model_name>/<string:_id>/' % base_path)

    for path, method in app.methods.items():
        resource = method.get_resource()
        resource.set_app(app)
        api.add_resource(resource, '%s/%s' % (base_path, path))

    if app.user_manager:
        umr = UserModelResource
        umr.set_app(app)

        uir = UserInstanceResource
        uir.set_app(app)

        api.add_resource(umr, '%s/user/' % app.crudcast_config['swagger']['basePath'])
        api.add_resource(uir, '%s/user/<string:_id>/' % app.crudcast_config['swagger']['basePath'])






import json


def is_jsonable(x):
    """
    Helper function to check that a variable can be JSON serialized.

    :param x: Variable to check
    :rtype: Boolean
    """
    try:
        json.dumps(x)
        return True
    except:
        return False


default_config = {
    # database
    "mongo_url": "mongodb://localhost:27017/",
    "db_name": "database",

    # routes
    'base_path': 'api',
}


# def get_config_option(key, default=default_config):
#     """
#     Simple helper function for retrieving config options. If option is populated in the YAML file, this option is
#     taken. Otherwise, it falls back to the options in the above `default_config`
#
#     :param key: the config option to be retrieved
#     :param default: the default config - defaults to the above `default_config` variable
#     """
#     return config.get('options', {}).get(key, default[key])


# client = pymongo.MongoClient(get_config_option('mongo_url'))
# db = client[get_config_option('db_name')]


# def get_config():
#     """
#     Returns the parsed config from the yaml file
#     """
#     return config


def get_models(config, db):
    """
    Populates the model list from the yaml config
    """
    models = {}

    for model_name, options in config['models'].items():
        m = {
            'name': model_name,
            'collection': db[model_name],
            'fields': options.get('fields', []),
        }
        models[model_name] = m

    return models



import base64
from flask import abort


class BaseAuthHandler(object):
    @classmethod
    def authenticate(cls, request, user):
        raise NotImplementedError()


class BasicAuth(object):
    @classmethod
    def authenticate(cls, request, user):
        auth_header = request.headers.get('Authorization')

        if not auth_header:
            abort(401)

        enc = auth_header.split()[1]
        decoded = base64.b64decode(enc.encode())
        username, password = decoded.decode().split(':')

        return user.authenticate(username, password)






from datetime import datetime
from crudcast.exceptions import ValidationError
from pymongo.collection import ObjectId
from bson.errors import InvalidId
from werkzeug.exceptions import NotFound


class BaseField(object):
    """
    Base class for field objects. All fields must extend this class

    :param name: Field name.
    :param model: the Model object to which the field belongs
    :type model: `models.Model`
    :param options: field options. The available options vary based on field type. Only the `required` and `unique`
                    options are implemented at this level

    """
    auto = False  # if an autofield, this value is `True`
    type = ''  # used by swagger

    def __init__(self, name, model, **options):
        self.name = name
        self.model = model
        self.required = options.get('required', False)
        self.unique = options.get('unique', False)

    def validate(self, data, _id=None):
        """
        When creating or updating an object. This method validates the inputs

        :param data: data to validate
        :type data: dict
        :param _id: ID as a string, or None in case of new objects
        :type _id: str or None

        :rtype: dict
        """
        if self.unique:
            query = {self.name: data}
            if _id:
                """
                If object is being updated, the query is adjusted to exclude the object being updated (otherwise, 
                it isn't possible to save an object without changing the unique field values every time)
                """
                query['_id'] = {'$not': {'$eq': ObjectId(_id)}}

            if self.model.collection.find(query).count() > 0:
                raise ValidationError('%s with this %s already exists' % (self.model.name, self.name),
                                      field=self.name)
        return data


class AutoBaseField(BaseField):
    """
    An automatically generated field. When `Model.create()` is called, the fields will be set automatically - the
    `.set()` method must be implemented in order for this to happen. Any attempt to set the value of this field
    manually will generate a ValidationError
    """
    auto = True

    def get_original(self, _id):
        """
        Helper function that returns the original value of the object, before it is updated. If you want to leave a
        value unchanged on update, the `.set()` method should call and return this function (otherwise the field
        value will be set to `None`)

        :param _id: the object id as a string
        """
        obj = self.model.to_repr(_id=ObjectId(_id))[0]
        return obj[self.name]

    def set(self, _id=None):
        """
        This method must be extended by all child classes. If it returns None, the value will not be changed

        :param _id: If an object is being updated, the _id will be provided
        """
        raise NotImplementedError()

    def validate(self, data, _id=None):
        """
        Any attempt to set the value of an auto field will raise a validation error
        """
        raise ValidationError('Auto fields cannot be set manually', field=self.name)


class StringField(BaseField):
    """
    String input field

    :param name: field name
    :param unique: if True, no new documents can be created if there is another document in the collection has the
                   same value for this field as the one being provided
    :type unique: Boolean
    """

    type = 'string'

    def __init__(self, name, **options):
        super(StringField, self).__init__(name, **options)

    def validate(self, data, _id=None):
        """
        Validates that the input value is a string
        """
        data = super().validate(data, _id=_id)
        try:
            assert isinstance(data, str)
        except AssertionError:
            raise ValidationError('Input must be a string', field=self.name)

        return data


class DateTimeField(BaseField):
    """
    A field for storing datetime objects

    :param name: field name
    :param format_string: format string to be used for validating input. Defaults to '%Y-%m-%d %H:%M:%S.%f' (ISO
                          format)

    """
    def __init__(self, name, **options):
        self.format_string = options.get('format_string', '%Y-%m-%d %H:%M:%S.%f')
        super(DateTimeField, self).__init__(name, **options)

    def validate(self, data, _id=None):
        """
        Validates the input string by asserting that the time format is valid
        """
        data = super().validate(data, _id=_id)
        try:
            datetime.strptime(data, self.format_string)
        except ValueError as err:
            raise ValidationError(err, field=self.name)
        return data


class NumberField(BaseField):
    """
    Numeric input field.
    """
    type = 'number'

    def validate(self, data, _id=None):
        """
        Asserts that the input is numeric
        """
        data = super().validate(data, _id=_id)
        try:
            assert isinstance(data, (int, float, complex))
        except AssertionError:
            raise ValidationError('Input must be numeric', field=self.name)

        return data


class AutoField(AutoBaseField):
    """
    An auto field, e.g. one that creates a sequential number, e.g. 1, 2, 3, etc. Cannot be set manually - fields that
    extend this must implement the `.set()` method
    """

    def set(self, _id=None):
        """
        Returns the collection length + 1 when the object is created
        """
        if _id:
            return self.get_original(_id)
        else:
            return self.model.collection.find({}).count() + 1


class AutoDateTimeField(AutoBaseField):
    def __init__(self, name, model, **options):
        """
        Automatically populates the current time at the moment of saving the object. If `create_only` is `True`,
        then the field only gets set when the document is created
        """
        super(AutoDateTimeField, self).__init__(name, model, **options)
        self.create_only = options.get('create_only', False)

    def set(self, _id=None):
        """
        Sets the field value to the current date/time when the object is created. If the object is being updated,
        and the value of `self.create_only` is `True`, then the date/time will be updated to the current time
        """
        if not self.create_only or not _id:
            return datetime.utcnow()
        else:
            return self.get_original(_id)


class BooleanField(BaseField):
    """
    A simple true/false field

    """
    type = 'boolean'

    def validate(self, data, _id=None):
        """
        Checks that the input value is a Boolean
        """
        data = super().validate(data, _id=_id)
        try:
            assert data in [True, False]
        except AssertionError:
            raise ValidationError('Input must be true or false', field=self.name)
        return data


class ForeignKeyField(BaseField):
    """
    A field that points at another collection in the database

    :param name: field name
    :param model: parent model
    :param to:  the related model's name
    """
    type = 'object'

    def get_related(self, related_model_name):
        """
        Returns the related model

        :param related_model_name: the name of another model in the app
        :rtype: crudcast.models.Model
        """
        from crudcast.models import Model
        return Model(name=related_model_name, app=self.model.app)

    def __init__(self, name, model, **options):
        super(ForeignKeyField, self).__init__(name, model, **options)
        self.related = self.get_related(options['to'])

    def validate(self, data, _id=None):
        """
        Checks to see if a document in the related model's collection matches the ID provided as `data`
        """
        data = super().validate(data, _id=_id)
        try:
            self.related.retrieve(data)
        except InvalidId:
            raise ValidationError('Invalid id', field=self.name)
        except NotFound:
            raise ValidationError('Cannot find %s with ID %s' % (self.related.name, data), field=self.name)

        return data


class ManyToManyField(ForeignKeyField):
    type = 'array'

    def validate(self, data, _id=None):
        """
        Checks to see if a document in the related model's collection matches the ID provided as `data`
        """
        for d in data:
            super().validate(d, _id=_id)

        return data



import importlib.util


class Method(object):
    def __init__(self, file, path, resource, **options):
        self.file = file
        self.path = path
        self.code = resource
        self.options = options

    def get_resource(self):
        _cls = self.code.split('.')[-1]
        spec = importlib.util.spec_from_file_location(self.code, self.file)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        return getattr(module, _cls)

    @property
    def swagger_definition(self):
        extra_swagger = self.options.get('swagger', {})

        res = self.get_resource()
        auth_type = self.options.get('auth_type')
        swagger = {}

        get = getattr(res, 'get', None)
        post = getattr(res, 'post', None)
        put = getattr(res, 'put', None)
        delete = getattr(res, 'delete', None)

        obj = {
                'tags': [self.code],
                'summary': 'Custom function',
                'consumes': 'application/json',
                'produces': 'application/json',
                'responses': {
                    '200': {
                        'description': 'successful operation',
                    }
                },
                'security': [
                    {
                        'basicAuth': []
                    }
                ] if auth_type else []
            }

        if get:
            swagger['get'] = obj
            for key, val in extra_swagger.get('get', {}).items():
                obj[key] = val
        if post:
            swagger['post'] = obj
            for key, val in extra_swagger.get('post', {}).items():
                obj[key] = val
        if put:
            swagger['put'] = obj
            for key, val in extra_swagger.get('put', {}).items():
                obj[key] = val
        if delete:
            swagger['delete'] = obj
            for key, val in extra_swagger.get('delete', {}).items():
                obj[key] = val

        return swagger





import pymongo
from yaml import load
from flask import Flask
from flask_swagger_ui import get_swaggerui_blueprint
from crudcast.models import Model
from crudcast.users import User
from crudcast.methods import Method
import os


class CrudcastApp(Flask):
    """
    The main Crudcast app object, which extends the Flask app

    :param import_name: The `import_name` parameter for Flask
    :param config_file: Local path to a valid `config.yml`
    :param kwargs: additional arguments to be passed to Flask
    """
    user_manager = None

    crudcast_config = {
        # database
        "mongo_url": "mongodb://localhost:27017/",
        "db_name": "database",

        # swagger
        'swagger': {
            'swagger': '2.0',
            'basePath': '/api',
            'url': '/api/docs',
            'info': {
                'description': 'This is an API automatically generated by crudcast',
                'version': '1.0.0',
                'title': 'My Crudcast app'
            }
        },
    } #: this is the default crudcast config. Any options supplied in your `config.yml` will overwrite these

    models = {}
    methods = {}
    user_config = {
        'fields': {},
        'options': {
            'auth_type': 'basic'
        }
    }

    client = None
    db = None

    def set_crudcast_config(self, config_file):
        """
        Overwrites `self.crudcast_config` with the content of the config file

        :param config_file: a local file path to a valid config file
        """
        with open(config_file, 'r') as f:
            options = load(f.read())

        for key, val in options.items():
            if key not in ['models', 'methods']:
                self.crudcast_config[key] = val

        self.client = pymongo.MongoClient(self.crudcast_config['mongo_url'])
        self.db = self.client[self.crudcast_config['db_name']]

        users = options.get('users')
        if not users:
            users = {}

        for key, val in users.items():
            self.user_config[key] = val

        if 'users' in options.keys():
            self.crudcast_config['users'] = users

        for model_name, model_options in options['models'].items():
            m = {
                'name': model_name,
                'collection': self.db[model_name],
                'fields': model_options.pop('fields', {}),
                'options': model_options
            }
            self.models[model_name] = m
            if 'users' in options.keys():
                self.models['user'] = {
                    'name': 'user',
                    'collection': self.db['user'],
                    'fields': self.user_config['fields'],
                    'options': self.user_config['options']
                }

        for method in options.get('methods', []):
            file = os.path.abspath(method.pop('file'))
            self.methods[method['path']] = Method(file=file, **method)

    def get_tag(self, model):
        """
        Returns the tag name/description to be used in the swagger view for each model

        :type model: crudcast.modles.Model
        :rtype: dict
        """
        return {
            'name': model.name,
            'description': model.options.get('description')
        }

    def get_model_path(self, model):
        """
        Generates a path entry for the swagger view for each model

        :type model: crudcast.models.Model
        :rtype: dict
        """
        parameters = []

        auth_type = model.options.get('auth_type')

        for field in model.fields:
            parameter = {
                'name': field.name,
                'in': 'query',
                'description': 'Filter for %s objects based on %s' % (model.name, field.name),
                'required': False,
            }

            parameters.append(parameter)

        post_parameters = [
            {
                'name': 'body',
                'in': 'body',
                'required': True,
                'description': 'New %s object' % model.name,
                'schema': {
                    '$ref': '#/definitions/%s' % model.name
                }
            }
        ]

        return {
            'get': {
                'tags': [model.name],
                'summary': 'List all %s objects' % model.name,
                'consumes': 'application/json',
                'produces': 'application/json',
                'parameters': parameters,
                'responses': {
                    '200': {
                        'description': 'successful operation',
                        'schema': {
                            '$ref': '#/definitions/%s' % model.name
                        }
                    }
                },
                'security': [
                    {
                        'basicAuth': []
                    }
                ] if auth_type else []
            },
            'post': {
                'tags': [model.name],
                'summary': 'Create a new %s' % model.name,
                'consumes': 'application/json',
                'produces': 'application/json',
                'parameters': post_parameters,
                'responses': {
                    '200': {
                        'description': 'object created',
                        'schema': {
                            '$ref': '#/definitions/%s' % model.name
                        }
                    }
                },
                'security': [
                    {
                        'basicAuth': []
                    }
                ] if auth_type else []
            }
        }

    def get_definition(self, model):
        """
        Returns the model definition based on the model's fields

        :type model: crudcast.models.Model
        :rtype: dict
        """
        definition_properties = {
            '_id': {
                'type': 'string',
            }
        }
        for field in model.fields:
            param = {
                'type': field.type
            }
            if field.type == 'array':
                param['items'] = {'type': 'string'}
            definition_properties[field.name] = param
        return {
            'type': 'object',
            'required': [field.name for field in model.fields if field.required],
            'properties': definition_properties
        }

    def get_instance_path(self, model):
        """
        Returns the path entries for instance level calls such as PUT, DELETE, RETRIEVE

        :type model: crudcast.models.Model
        :rtype: dict
        """
        parameters = [
            {
                'name': '_id',
                'in': 'path',
                'required': True,
                'type': 'string',
                'description': 'ID of %s object' % model.name
            }
        ]
        post_parameters = [
            {
                'name': '_id',
                'in': 'path',
                'required': True,
                'type': 'string',
                'description': 'ID of %s object' % model.name
            },
            {
                'name': 'body',
                'in': 'body',
                'required': True,
                'description': 'New %s object' % model.name,
                'schema': {
                    '$ref': '#/definitions/%s' % model.name
                }
            }
        ]
        return {
            'get': {
                'tags': [model.name],
                'summary': 'Retrieve a %s object' % model.name,
                'parameters': parameters,
                'consumes': 'application/json',
                'produces': 'application/json',
                'responses': {
                    '200': {
                        'description': 'successful operation',
                        'schema': {
                            '$ref': '#/definitions/%s' % model.name
                        }
                    }
                }
            },
            'put': {
                'tags': [model.name],
                'summary': 'Update a %s object' % model.name,
                'consumes': 'application/json',
                'produces': 'application/json',
                'parameters': post_parameters,
                'responses': {
                    '200': {
                        'description': 'successful operation',
                        'schema': {
                            '$ref': '#/definitions/%s' % model.name
                        }
                    }
                }
            },
            'delete': {
                'tags': [model.name],
                'summary': 'Delete a %s object' % model.name,
                'parameters': parameters,
                'consumes': 'application/json',
                'produces': 'application/json',
                'responses': {
                    '200': {
                        'description': 'successful operation',
                        'schema': {
                            '$ref': '#/definitions/%s' % model.name
                        }
                    }
                }
            }
        }

    def get_security_definitions(self):
        return {
                'basicAuth': {
                    'type': 'basic',
                }
        }

    @property
    def swagger_config(self):
        """
        Retrieves the complete swagger configuration

        :rtype: dict
        """
        config = self.crudcast_config['swagger']
        tags = []
        paths = {}
        definitions = {}

        for model_name, model_attrs in self.models.items():
            model = Model(model_name, self)
            tags.append(self.get_tag(model))
            paths['/%s/' % model_name] = self.get_model_path(model)
            paths['/%s/{_id}/' % model_name] = self.get_instance_path(model)
            definitions[model.name] = self.get_definition(model)

        for method_path, method in self.methods.items():
            paths['/' + method_path] = method.swagger_definition

        config['tags'] = tags
        config['paths'] = paths
        config['definitions'] = definitions

        if self.user_manager:
            config['securityDefinitions'] = self.get_security_definitions()

        return config

    def get_swagger_ui_view(self):
        """
        Creates a swagger view using `flask_swagger_ui`

        :return: a swagger view
        """
        return get_swaggerui_blueprint(
            self.crudcast_config['swagger']['url'],
            '/swagger',
        )

    def __init__(self, import_name, config_file, **kwargs):
        self.set_crudcast_config(config_file)

        if 'users' in self.crudcast_config.keys():
            self.user_manager = User(app=self, **self.crudcast_config['users'])

        super().__init__(import_name, **kwargs)






from crudcast.api import get_api
import argparse
from crudcast.app import CrudcastApp
from crudcast.exceptions import ValidationError, handle_invalid_usage
import json
import getpass


SWAGGER_URL = '/api/docs'  # URL for exposing Swagger UI (without trailing '/')


def main():
    parser = argparse.ArgumentParser(description='Process some integers.')
    cf = parser.add_argument('--config-file', help='Path to yml config file', dest='config_file', default='config.yml')
    parser.add_argument('--create-user', dest='create_admin', default=False, action='store_true')
    parser.add_argument('--host', help='Host name', dest='host', default='0.0.0.0')
    parser.add_argument('--port', help='Port number', dest='port', default=5000)
    parser.add_argument('--no-load-dotenv', help='Disable dotenv', dest='no_load_dotenv', default=False,
                        action='store_true')
    parser.add_argument('--debug', help='Debug mode', dest='debug', default=False, type=bool)
    parser.add_argument('--import-name', help='Flask app import name', dest='import_name', default='Crudcast')

    args = parser.parse_args()
    create_admin = args.create_admin

    config_file = args.config_file
    import_name = args.import_name
    port = args.port

    debug = args.debug
    host = args.host

    try:
        app = CrudcastApp(import_name=import_name, config_file=config_file)
    except FileNotFoundError:
        raise argparse.ArgumentError(cf, 'Config file not found: %s' % config_file)

    if create_admin:
        user_manager = app.user_manager
        username = input('Enter username: ')
        password = getpass.getpass('Enter password: ')
        password_confirm = getpass.getpass('Confirm password: ')
        try:
            assert password == password_confirm
        except AssertionError:
            raise AssertionError('Passwords do not match')

        user_manager.create({'username': username, 'password': password})
        return 'Created user %s' % username

    else:
        load_dotenv = not args.no_load_dotenv

        get_api(app)
        app.register_error_handler(ValidationError, handle_invalid_usage)

        app.register_blueprint(app.get_swagger_ui_view(), url_prefix=SWAGGER_URL)

        @app.route('/swagger')
        def swagger_file():
            return json.dumps(app.swagger_config)

        app.run(host=host, port=port, debug=debug, load_dotenv=load_dotenv)




import os
from setuptools import find_packages, setup
from crudcast import __version__

os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))
name = 'crudcast'
version = __version__

setup(
    name=name,
    version=version,
    packages=find_packages(),
    package_data={'crudcast': ['requirements.txt']},
    include_package_data=True,
    license='Apache Software License',
    long_description='',
    author='Christopher Davies',
    author_email='christopherdavies553@gmail.com',
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
    ],
    install_requires=[
        'Click==7.0',
        'Flask==1.0.2',
        'itsdangerous==1.1.0',
        'Jinja2==2.10',
        'MarkupSafe==1.1.0',
        'pymongo==3.7.2',
        'PyYAML==3.13',
        'Werkzeug==0.14.1',
        'flask_swagger_ui==3.18.0',
        'flask_restplus==0.12.1',
        'sphinx_press_theme==0.1.1',
        'passlib==1.7.1',
        'bcrypt==3.1.5'
    ],
    entry_points={
        'console_scripts': ['crudcast=crudcast.entrypoint:main'],
    }
)



from crudcast.utils import is_jsonable
from crudcast.exceptions import ValidationError
from pymongo.collection import ObjectId
from crudcast.fields import (
    StringField, NumberField, DateTimeField, BooleanField, ForeignKeyField, AutoField, AutoDateTimeField,
    ManyToManyField
)
from flask import abort
from crudcast.authentication import BasicAuth


class Model(object):
    def __init__(self, name, app):
        """
        A model object - effectively a schema for the mongodb

        :param name: model name
        """
        self.name = name
        self.app = app

        try:
            self.object = self.app.models[name]
        except KeyError:
            abort(404)
        self.collection = self.object['collection']
        self.fields = self.set_fields(self.object['fields'])
        self.options = self.object.get('options', {})

    def set_fields(self, fields):
        """
        Sets the list of fields
        """
        cleaned_fields = []
        field_types = {
            'string': StringField,
            'number': NumberField,
            'datetime': DateTimeField,
            'boolean': BooleanField,
            'foreignkey': ForeignKeyField,
            'autofield': AutoField,
            'auto_datetime': AutoDateTimeField,
            'manytomany': ManyToManyField,
        }

        for field_name, options in fields.items():
            if not options:
                options = {}

            field_type = field_types[options.get('type', 'string')]

            field = field_type(name=field_name, model=self, **options)

            cleaned_fields.append(field)

        return cleaned_fields

    @property
    def fieldnames(self):
        return [field.name for field in self.fields]

    def get_field_by_name(self, key):
        return list(filter(lambda x: x.name == key, self.fields))[0]

    def validate(self, data=None, _id=None):
        # prevent 500 error on empty payload
        if not data:
            data = {}

        # ensure that required fields are supplied, if creating
        if not _id:
            required_fields = [field.name for field in self.fields if field.required]
            for f in required_fields:
                try:
                    assert f in data.keys()
                except AssertionError:
                    raise ValidationError('This field is required', field=f)

        # ensure that all input fields are valid field names
        for key, val in data.items():
            if not key in self.fieldnames:
                raise ValidationError('Invalid field', field=key)

            # perform field-level validation
            field = self.get_field_by_name(key)
            field.validate(val, _id=_id)

        auto_fields = [field for field in self.fields if field.auto]

        for f in auto_fields:
            data[f.name] = f.set(_id=_id)

        return data

    def find(self, **query):
        q = {}
        for key, val in query.items():
            try:
                q[key] = ''.join(val)
            except TypeError:
                q[key] = val
        return self.collection.find(q)

    def to_repr(self, **query):
        items = self.find(**query)
        response = []
        for item in items:
            obj = {}
            for key, val in item.items():
                if is_jsonable(val):
                    obj[key] = val
                else:
                    obj[key] = str(val)

            response.append(obj)

        return response

    def retrieve(self, _id):
        if self.exists(_id):
            return self.to_repr(_id=ObjectId(_id))[0]
        else:
            abort(404)

    def create(self, data):
        data = self.validate(data)
        obj = self.collection.insert_one(data)
        return self.retrieve(obj.inserted_id)

    def update(self, _id, data):
        if not self.exists(_id):
            abort(404)

        data = self.validate(data, _id=_id)
        self.collection.find_one_and_update({'_id': ObjectId(_id)}, {'$set': data})
        return self.retrieve(_id)

    def exists(self, _id):
        """
        Check to see if a document with a given ID exists

        :param _id: the id to check
        :type _id: str
        :rtype: bool
        """
        return self.collection.find({'_id': ObjectId(_id)}).count() == 1

    def delete(self, _id):
        if self.exists(_id):
            self.collection.find_one_and_delete({'_id': ObjectId(_id)})
            return {}
        else:
            abort(404)

    def get_auth_type(self):
        """
        Returns an auth class for the model, if any
        """
        mappings = {
            'basic': BasicAuth
        }
        auth_type = self.app.models[self.name]['options'].get('auth_type')
        if auth_type:
            return mappings.get(auth_type)



import unittest
import mock
from mocks import MockApi, MockApp, MockParser, MockModel
from fields import (
    BaseField, AutoBaseField, StringField, DateTimeField, NumberField, AutoField, AutoDateTimeField, BooleanField,
    ForeignKeyField
)
from models import Model
from api import get_api
from app import CrudcastApp
import json
from datetime import datetime


crudcast_config = {
    'models': {
        'test': {
            'fields': {
                'm2m': {
                    'type': 'manytomany',
                    'to': 'test2'
                },
                'name': {
                    'type': 'string',
                    'unique': True,
                    'required': True
                }
            }
        },
        'test2': {}
    },
    'something_else': None
}


class CrudCastTestCase(unittest.TestCase):
    @mock.patch('flask_restplus.Api', MockApi)
    def test_api(self, *args):
        get_api(MockApp)

    def test_app(self, *args):
        with mock.patch("builtins.open", mock.mock_open(read_data=json.dumps(crudcast_config))):
            app = CrudcastApp(__name__, config_file='file')
            self.assertEqual(app.swagger_config['swagger'], '2.0')

            with mock.patch('flask_swagger_ui.get_swaggerui_blueprint'):
                app.get_swagger_ui_view()

    @mock.patch('crudcast.app.CrudcastApp', MockApp)
    @mock.patch('argparse.ArgumentParser', MockParser)
    def test_entrypoint(self, *args):
        with mock.patch("builtins.open", mock.mock_open(read_data=json.dumps(crudcast_config))):
            from entrypoint import main
            main()

    def test_exceptions(self, *args,):
        from crudcast.exceptions import handle_invalid_usage, ValidationError
        with self.assertRaises(RuntimeError):
            handle_invalid_usage(ValidationError('test', status_code=415))

    def test_fields(self):
        f = BaseField(name='test', model=MockModel(), unique=True)
        data = f.validate(data=None, _id='507f1f77bcf86cd799439011')
        self.assertIsNone(data)

        model = MockModel(count=10)

        with self.assertRaises(Exception):
            f = BaseField(name='test', model=model, unique=True)
            f.validate(data=None, _id='507f1f77bcf86cd799439011')

        f = AutoBaseField(name='test', model=MockModel())
        f.get_original('507f1f77bcf86cd799439011')

        with self.assertRaises(NotImplementedError):
            f.set(None)

        with self.assertRaises(Exception):
            f.validate(None)

        s = StringField('test', model=MockModel())
        data = s.validate('test')
        self.assertEqual(data, 'test')

        with self.assertRaises(Exception):
            s.validate(None)

        dt = DateTimeField('test', model=MockModel())
        dt.validate('2018-12-10 15:00:00.123')

        with self.assertRaises(Exception):
            dt.validate('invalid')

        nf = NumberField('test', model=MockModel())
        data = nf.validate(10)
        self.assertEqual(data, 10)

        with self.assertRaises(Exception):
            nf.validate('test')

        af = AutoField('test', model=MockModel())
        val = af.set()
        self.assertEqual(val, 1)

        self.assertEqual(af.set(_id='507f1f77bcf86cd799439011'), {})

        adt = AutoDateTimeField('test', model=MockModel(),create_only=True)
        today = datetime.now().today()

        time = adt.set()
        self.assertEqual(today.date(), time.date())

        new_time = adt.set(_id='507f1f77bcf86cd799439011')
        self.assertEqual(new_time, {})

        bf = BooleanField('test', model=MockModel())
        data = bf.validate(True)
        self.assertTrue(data)

        with self.assertRaises(Exception):
            bf.validate('test')

        fk = ForeignKeyField('test', model=MockModel(), to='test')
        with self.assertRaises(Exception):
            fk.validate('507f1f77bcf86cd799439011')

    def test_model(self):
        app = MockApp()
        model = Model('test', app=app)
        self.assertEqual([], model.fieldnames)
        with self.assertRaises(IndexError):
            self.assertEqual([], model.get_field_by_name('test'))

        model.validate({})
        model.find(test='test', test_two=['test'], test_three=False)
        model.to_repr()

        with self.assertRaises(Exception):
            model.create({})

        with self.assertRaises(Exception):
            model.update(_id='507f1f77bcf86cd799439011', data={})

    def test_methods(self):
        from crudcast.methods import Method
        from crudcast.resources import Resource
        import os
        root = os.path.abspath(os.path.join(__file__, os.pardir))
        file = os.path.join(root, 'mocks.py')

        method = Method(file, '/test/<string:arg1>', 'mocks.TestResource')
        _cls = method.get_resource()
        self.assertTrue(issubclass(_cls, Resource))



from flask import request
from flask_restplus import Resource as BaseResource
from crudcast.models import Model
from crudcast.users import User


class Resource(BaseResource):
    """
    Extension to the default `flask_restplus.BaseResource` which sets the `app` paramter, which is used for
    model creation
    """
    app = None
    model = None

    def check_auth(self):
        auth_type = self.model.get_auth_type()
        if auth_type:
            return auth_type.authenticate(request=request, user=self.app.user_manager)

    @classmethod
    def set_app(cls, app):
        """
        Sets the value of __class__.app

        :type app: crudcast.app.CrudcastApp
        """
        cls.app = app


class ModelResource(Resource):
    """
    Model level resources - implements all REST methods for paths with no ID
    """
    def get(self, model_name):
        """
        Lists all instances of a given model

        :param model_name: the name of the model, as it appears in the config file
        :return: a list of instances of the model object
        """
        self.model = Model(model_name, self.app)
        user = self.check_auth()
        return self.model.to_repr(**request.args)

    def post(self, model_name):
        """
        Create an instance of a model

        :param model_name: the name of the model, as it appears in the config file
        :return: details of the created instance
        """
        self.model = Model(model_name, self.app)
        user = self.check_auth()
        response = self.model.create(request.json)
        return response


class InstanceResource(Resource):
    """
    Instance level resources - implements all REST methods for paths with an ID
    """
    def get(self, model_name, _id):
        """
        Retrieve a single instance by its ID

        :param model_name: the name of the model, as it appears in the config file
        :param _id: MongoDB _id string
        :return: the MongoDB document
        """
        self.model = Model(model_name, self.app)
        user = self.check_auth()
        instance = self.model.retrieve(_id)
        return instance

    def put(self, model_name, _id):
        """
        Update a single instance by its ID

        :param model_name: the name of the model, as it appears in the config file
        :param _id: MongoDB _id string
        :return: the MongoDB document
        """

        self.model = Model(model_name, self.app)
        user = self.check_auth()
        instance = self.model.update(_id=_id, data=request.json)
        return instance

    def delete(self, model_name, _id):
        """
        Delete a single instance by its ID

        :param model_name: the name of the model, as it appears in the config file
        :param _id: MongoDB _id string
        """

        self.model = Model(model_name, self.app)
        user = self.check_auth()
        instance = self.model.delete(_id=_id)
        return instance


class UserModelResource(Resource):
    def get(self):
        """
        Lists all users

        :return: a list of instances of the user object
        """
        self.model = User(self.app)
        user = self.check_auth()
        return self.model.to_repr(**request.args)

    def post(self):
        """
        Create a user

        :return: details of the created user
        """
        self.model = User(self.app)
        user = self.check_auth()
        return self.model.create(request.json)


class UserInstanceResource(Resource):
    """
    Instance level resources - implements all REST methods for paths with an ID
    """
    def get(self, _id):
        """
        Retrieve a single instance by its ID

        :param _id: MongoDB _id string
        :return: the MongoDB document
        """
        self.model = User(self.app)
        print('manager', self.app.user_manager)
        user = self.check_auth()
        return self.model.retrieve(_id)

    def put(self, _id):
        """
        Update a single instance by its ID

        :param model_name: the name of the model, as it appears in the config file
        :param _id: MongoDB _id string
        :return: the MongoDB document
        """

        self.model = User(self.app)
        user = self.check_auth()
        return self.model.update(_id=_id, data=request.json)

    def delete(self, _id):
        """
        Delete a single instance by its ID

        :param _id: MongoDB _id string
        """

        self.model = User(self.app)
        user = self.check_auth()
        return self.model.delete(_id=_id)
